#include <string.h>
#include <malloc.h>
#include <ctype.h>
#include <unixodbc.h>
#include "sybase.h"

#define EX_SQL_C_TCHAR SQL_C_TCHAR
#define EX_SQLTCHAR SQLTCHAR
#define _DBTEXT(s) (EX_SQLTCHAR ODBCFAR *)(s)
#define SQL_MAX_NAME_LEN 200

typedef struct
{
 EX_SQLTCHAR *value;
 SDWORD size;
 EX_SQLTCHAR *colname;
 unsigned int width;
} columns;

columns  *Columns = NULL;
SWORD    NumColumns;
HENV     Environment;
HDBC     Connection;
HSTMT    Statement;

EX_SQLTCHAR error_msg[300];

char *get_error(void)
{
 return((char *)error_msg);
}

static int allokay(RETCODE rcode, HSTMT stmt)
{
 EX_SQLTCHAR sqlstate[6];

 if(rcode == SQL_SUCCESS || rcode == SQL_SUCCESS_WITH_INFO)
  return(OK);
 else
  if(rcode == SQL_NO_DATA_FOUND)
	return(ERROR);
  else
  {
	SQLError(Environment, Connection, stmt, sqlstate, NULL,
            error_msg, sizeof(error_msg)/sizeof(error_msg[0]), NULL);
	return(ERROR);
  }
}

int make_columns(HSTMT statement)
{
 EX_SQLTCHAR colname[ NAME_LEN + 1 ];
 SWORD       namelen;
 UWORD       col;
 SDWORD      size;

 if(!allokay(SQLNumResultCols(statement, &NumColumns), statement))
  return(ERROR);
 Columns = (columns *) malloc(NumColumns * sizeof(columns));
 memset(Columns, 0, NumColumns * sizeof(columns));
 for(col = 0; col < NumColumns; ++col)
 {
  if(!allokay(SQLColAttributes(statement, (UWORD) (col + 1),
              SQL_COLUMN_DISPLAY_SIZE, NULL, 0, NULL, &size), statement))
   return(ERROR);
  if(size > MAX_FETCH_SIZE)
   size = MAX_FETCH_SIZE;
  Columns[col].width = (int) size;
  Columns[col].value = (EX_SQLTCHAR *) malloc((int)(size + 1)
							                          * sizeof(EX_SQLTCHAR));
  if(!allokay(SQLBindCol(statement, (UWORD) (col + 1), EX_SQL_C_TCHAR,
              Columns[col].value, (size + 1) * sizeof(EX_SQLTCHAR),
              &Columns[col].size), statement))
   return(ERROR);
  if(!allokay(SQLColAttributes(statement, (UWORD) (col + 1),
              SQL_COLUMN_NAME, colname, NAME_LEN * sizeof(EX_SQLTCHAR),
              &namelen, NULL), statement))
   return(ERROR);
  Columns[col].colname = (EX_SQLTCHAR *) malloc(namelen + sizeof(EX_SQLTCHAR));
  strcpy((char *) Columns[col].colname, (char *) colname);
  if(Columns[col].width < (unsigned) namelen)
   Columns[col].width = namelen;
  if(Columns[col].width < NULL_TEXT_LEN )
   Columns[col].width = NULL_TEXT_LEN;
 }
 return(OK);
}

static void free_columns()
{
 int col;

 for(col = 0; col < NumColumns; ++col)
 {
  free(Columns[col].value);
  free(Columns[col].colname);
 }
 free(Columns);
 Columns = NULL;
 NumColumns = 0;
}

int exec_proc(char *buff, int *param)
{
 long buf;

 if(!allokay(SQLAllocStmt(Connection, &Statement), NULL))
  return(ERROR);
 if(!allokay(SQLBindParameter(Statement, 1, SQL_PARAM_OUTPUT, SQL_C_LONG, SQL_INTEGER, 0, 0, param, 0, &buf), Statement))
  return(ERROR);
 if(!allokay(SQLExecDirect(Statement, buff, SQL_NTS), Statement))
  return(ERROR);
 return(OK);
}

int open_cursor(char *buff)
{
 if(!allokay(SQLAllocStmt(Connection, &Statement), NULL))
  return(ERROR);
 if(!allokay(SQLSetScrollOptions(Statement, SQL_CONCUR_VALUES, SQL_SCROLL_DYNAMIC, 1), Statement))
  return(ERROR);
 if(!allokay(SQLPrepare(Statement, (EX_SQLTCHAR ODBCFAR *)buff, SQL_NTS), Statement))
  return(ERROR);
 if(!allokay(SQLExecute(Statement), Statement))
  return(ERROR);
 make_columns(Statement);
 return(OK);
}

int close_cursor()
{
 free_columns();
 SQLFreeStmt(Statement, SQL_DROP);
 Statement = NULL;
 return(OK);
}

int fetch_next()
{
 int okay;
 UDWORD numfetch;

 okay = allokay(SQLExtendedFetch(Statement, SQL_FETCH_NEXT, 0,
                &numfetch, NULL), Statement);
 if(!okay)
  return(ERROR);
 if(numfetch == 0)
 {
  strcpy(error_msg, "Not fount while fetching next");
  return(ERROR);
 }
 return(OK);
}

int fetch_relative(long relpos)
{
 UDWORD numfetch;
 int okay;

 if(relpos == 0)
  return(OK);
 okay = allokay(SQLExtendedFetch(Statement, SQL_FETCH_RELATIVE,
			relpos, &numfetch, NULL), Statement);
 if(!okay)
  return(ERROR);
 if(numfetch == 0)
 {
  sprintf(error_msg, "Not fount while fetching %d", relpos);
  return(ERROR);
 }
 return(OK);
}

int fetch_first(void)
{
 UDWORD numfetch;
 int    okay;

 okay = allokay(SQLExtendedFetch(Statement, SQL_FETCH_FIRST, 0,
                &numfetch, NULL), Statement);
 if(okay)
  okay = allokay(SQLExtendedFetch(Statement, SQL_FETCH_RELATIVE,
                 -1, &numfetch, NULL), Statement);

 return(okay);
}

int fetch_last(void)
{
 UDWORD numfetch;

 return(allokay(SQLExtendedFetch(Statement, SQL_FETCH_LAST, 0,
                 &numfetch, NULL), Statement));
}

int get_total(void)
{
 return NumColumns;
}

unsigned int get_width(int fieldnum)
{
 return(Columns[fieldnum].width);
}

char *get_name(int fieldnum)
{
 return((char *)Columns[fieldnum].colname);
}

char *get_value(int fieldnum)
{
 if(Columns[fieldnum].size == SQL_NULL_DATA)
  return(NULL_TEXT);
 else
  return((char *)Columns[fieldnum].value);
}

int do_request(char *buff)
{
 if(!allokay(SQLAllocStmt(Connection, &Statement), Statement))
  return(ERROR);
 if(!allokay(SQLExecDirect(Statement, buff, SQL_NTS), Statement))
  return(ERROR);
 if(!allokay(SQLFreeStmt(Statement, SQL_DROP), Statement))
  return(ERROR);
 Statement = NULL;

 return(OK);
}

int odbc_init(char *basename)
{
 if(SQLAllocEnv(&Environment) != SQL_SUCCESS)
  return(ERROR);
 if(SQLAllocConnect(Environment, &Connection) != SQL_SUCCESS)
 {
  SQLFreeEnv(Environment);
  return(ERROR);
 }
 if(!allokay(SQLConnect(Connection, _DBTEXT(basename), SQL_NTS,
                        _DBTEXT("DBA"), SQL_NTS, _DBTEXT("SQL"), SQL_NTS ), NULL))
 {
  strcpy(error_msg, "Unable to open data source");
  return(ERROR);
 }
 return(OK);
}

int odbc_close()
{
 close_cursor();
 SQLTransact( Environment, Connection, SQL_ROLLBACK );
 SQLDisconnect( Connection );
 SQLFreeConnect( Connection );
 SQLFreeEnv( Environment );
 return(OK);
}
